
uniform sampler2D TransmittanceTexture;

// Parameterization functions

vec2 getTransmittanceUV(in float r, in float mu) {
	float uR = sqrt((r - Rg) / (Rt - Rg));
	float uMu = atan((mu + 0.15) / (1.0 + 0.15) * tan(1.5)) / 1.5;
    return vec2(uMu, uR);
}

void getTransmittanceRMu(out float r, out float muS) {
	r = gl_FragCoord.y / float(TRANSMITTANCE_H);
	muS = gl_FragCoord.x / float(TRANSMITTANCE_W);
	r = Rg + (r * r) * (Rt - Rg);
	muS = -0.15 + tan(1.5 * muS) / tan(1.5) * (1.0 + 0.15);
}

vec2 getIrradianceUV(in float r, in float muS) {
	float uR = (r - Rg) / (Rt - Rg);
	float uMuS = (muS + 0.2) / (1.0 + 0.2);
	return vec2(uMuS, uR);
}

void getIrradianceRMuS(out float r, out float muS) {
	r = Rg + (gl_FragCoord.y - 0.5) / (float(SKY_H) - 1.0) * (Rt - Rg);
	muS = -0.2 + (gl_FragCoord.x - 0.5) / (float(SKY_W) - 1.0) * (1.0 + 0.2);
}

vec4 texture4D(in sampler3D table, in float r, in float mu, in float muS, in float nu) {
	float H = sqrt(Rt * Rt - Rg * Rg);
	float rho_sqr = r * r - Rg * Rg;
	float rho = sqrt(rho_sqr);
	float rmu = r * mu;
	float delta = rmu * rmu - rho_sqr;
	
	// calculate parameters uNu, uMuS, uMu and uR and fit to texture resolution
	
	float uNu = (1.0 + nu) / 2.0;
	
	float uNuRes = uNu * (float(RES_NU) - 1.0);
	float uNuResMin = floor(uNuRes);
	
	float uNuDiff = uNuRes - uNuResMin;
	
	float uMuS = (atan(max(muS, -0.1975) * tan(1.26 * 1.1)) / 1.1 + (1.0 - 0.26)) * 0.5;
	float uMuSRes = 0.5 / float(RES_MU_S) + uMuS * (1.0 - 1.0 / float(RES_MU_S));
	
	float uMu, uMuRes;
	if (rmu < 0.0 && delta > 0.0) {
		uMu = (rmu + sqrt(delta)) / (2.0 * rho);
		uMuRes = 0.5 - 0.5 / float(RES_MU) + uMu * (1.0 - 1.0 / float(RES_MU));
	}
	else {
		uMu = -(rmu - sqrt(delta + H * H)) / (2.0 * (rho + H));
		uMuRes = 0.5 + 0.5 / float(RES_MU) + uMu * (1.0 - 1.0 / float(RES_MU));
	}
	
	float uR = rho / H;
	float uRRes = 0.5 / float(RES_R) + uR * (1.0 - 1.0 / float(RES_R));
	
	// final lookup coords
	vec3 uvw1 = vec3((uNuResMin + uMuSRes) / float(RES_NU), uMuRes, uRRes);
	vec3 uvw2 = vec3((uNuResMin + 1.0 + uMuSRes) / float(RES_NU), uMuRes, uRRes);
	
	vec4 value1 = texture(table, uvw1);
	vec4 value2 = texture(table, uvw2);
	
	// interpolate between values at different nu
	return mix(value1, value2, uNuDiff);
}

void getMuMuSNu(in float r, in vec4 dhdH, out float mu, out float muS, out float nu) {
	float x = gl_FragCoord.x - 0.5;
	float y = gl_FragCoord.y - 0.5;
	
	if (y < float(RES_MU) / 2.0) {
		float d = 1.0 - y / (float(RES_MU) / 2.0 - 1.0);
		d = min(max(dhdH.z, d * dhdH.w), dhdH.w * 0.999);
		mu = (Rg * Rg - r * r - d * d) / (2.0 * r * d);
		mu = min(mu, -sqrt(1.0 - (Rg / r) * (Rg / r)) - 0.001);
	} else {
		float d = (y - float(RES_MU) / 2.0) / (float(RES_MU) / 2.0 - 1.0);
		d = min(max(dhdH.x, d * dhdH.y), dhdH.y * 0.999);
		mu = (Rt * Rt - r * r - d * d) / (2.0 * r * d);
	}
	muS = mod(x, float(RES_MU_S)) / (float(RES_MU_S) - 1.0);
	muS = tan((2.0 * muS - 1.0 + 0.26) * 1.1) / tan(1.26 * 1.1);
	nu = -1.0 + floor(x / float(RES_MU_S)) / (float(RES_NU) - 1.0) * 2.0;
}

// Utility functions

// nearest intersection of ray r,mu with ground or top atmosphere boundary
// mu=cos(ray zenith angle at ray origin)
float limit(in float r, in float mu) {
	float dout = -r * mu + sqrt(r * r * (mu * mu - 1.0) + RL * RL);
	float delta2 = r * r * (mu * mu - 1.0) + Rg * Rg;
	if (delta2 >= 0.0) {
		float din = -r * mu - sqrt(delta2);
		if (din >= 0.0) {
			dout = min(dout, din);
		}
	}
	return dout;
}

// transmittance(=transparency) of atmosphere for infinite ray (r,mu)
// (mu=cos(view zenith angle)), intersections with ground ignored
vec3 transmittance(in float r, in float mu) {
	vec2 uv = getTransmittanceUV(r, mu);
    return texture(TransmittanceTexture, uv).rgb;
}

// transmittance(=transparency) of atmosphere for infinite ray (r,mu)
// (mu=cos(view zenith angle)), or zero if ray intersects ground
vec3 transmittanceWithShadow(in float r, in float mu) {
	if (mu < -sqrt(1.0 - (Rg / r) * (Rg / r))) {
		return vec3(0.0);
	} else {
		return transmittance(r, mu);
	}
}

// transmittance(=transparency) of atmosphere between x and x0
// assume segment x,x0 not intersecting ground
// r=||x||, mu=cos(zenith angle of [x,x0) ray at x), v=unit direction vector of [x,x0) ray
vec3 transmittance(in float r, in float mu, in vec3 v, in vec3 x0) {
	float r1 = length(x0);
	float mu1 = dot(x0, v) / r;

	vec4 params1 = vec4(r, mu, r1, mu1);
	vec4 params2 = vec4(r1, -mu1, r, -mu);

	vec4 p = mix(params1, params2, step(0.0, -mu));
	return min(transmittance(p.x, p.y) / transmittance(p.z, p.w), 1.0);
}

// optical depth for ray (r,mu) of length d, using analytic formula
// (mu=cos(view zenith angle)), intersections with ground ignored
// H=height scale of exponential density function
float opticalDepth(in float H, in float r, in float mu, in float d) {
	float a = sqrt((0.5 / H) * r);
	vec2 a01 = a * vec2(mu, mu + d / r);
	vec2 a01s = sign(a01);
	vec2 a01sq = a01 * a01;
	float x = a01s.y > a01s.x ? exp(a01sq.x) : 0.0;
	vec2 y = a01s / (2.3193 * abs(a01) + sqrt(1.52 * a01sq + 4.0)) * vec2(1.0, exp(-d / H * (d / (2.0 * r) + mu)));
	return sqrt((6.2831 * H) * r) * exp((Rg - r) / H) * (x + dot(y, vec2(1.0, -1.0)));
}

// transmittance(=transparency) of atmosphere for ray (r,mu) of length d
// (mu=cos(view zenith angle)), intersections with ground ignored
// uses analytic formula instead of transmittance texture
vec3 analyticTransmittance(in float r, in float mu, in float d) {
	return exp(- betaR * opticalDepth(HR, r, mu, d) - betaMEx * opticalDepth(HM, r, mu, d));
}

// transmittance(=transparency) of atmosphere between x and x0
// assume segment x,x0 not intersecting ground
// d = distance between x and x0, mu=cos(zenith angle of [x,x0) ray at x)
vec3 transmittance(in float r, in float mu, in float d) {
	float r1 = sqrt(r * r + d * d + 2.0 * r * mu * d);
	
	float mu1 = (r * mu + d) / r1;

	vec4 params1 = vec4(r, mu, r1, mu1);
	vec4 params2 = vec4(r1, -mu1, r, -mu);

	vec4 p = mix(params1, params2, step(0.0, -mu));
	return min(transmittance(p.x, p.y) / transmittance(p.z, p.w), 1.0);
}

vec3 irradiance(in sampler2D sampler, in float r, in float muS) {
	vec2 uv = getIrradianceUV(r, muS);
	return texture(sampler, uv).rgb;
}

// Rayleigh phase function
float phaseFunctionR(in float mu) {
	return (3.0 / (16.0 * M_PI)) * (1.0 + mu * mu);
}

// Mie phase function
float phaseFunctionM(in float mu) {
	return 1.5 * 1.0 / (4.0 * M_PI) * (1.0 - mieG*mieG) * pow(1.0 + (mieG*mieG) - 2.0*mieG*mu, -3.0/2.0) * (1.0 + mu * mu) / (2.0 + mieG*mieG);
}

// approximated single Mie scattering (cf. approximate Cm in paragraph "Angular precision")
vec3 getMie(in vec4 rayMie) { // rayMie.rgb=C*, rayMie.w=Cm,r
	return rayMie.rgb * rayMie.w / max(rayMie.r, 1e-4) * (betaR.r / betaR);
}
