
noperspective in vec3 ray;

uniform vec2 WindowSize;
uniform mat4 InvProjMatrix, InvModelViewMatrix;

// todo: make uniform
const float ISun = 100.0;
const float exposure = 0.4;

uniform vec3 cam_pos, sun_dir, vs_sun_dir;

uniform sampler3D InscatterTexture;
uniform sampler2D IrradianceTexture;

uniform sampler2D InputTexture;

vec3 attenuation(in float r, in float mu, in float t) {
	return analyticTransmittance(r, mu, t);
}

#define FIX

// inscattered light along ray x+tv, when sun in direction s (=S[L]-T(x,x0)S[L]|x0)
vec3 ground_inscatter(in vec3 x, in vec3 x0, in float t, in vec3 v, in vec3 s, in float r, in float mu, in vec3 attenuation) {
	// Assume x is not in space, and ray intersects the atmosphere and the ground
	
	float nu = dot(v, s);
	float muS = dot(x, s) / r;
	float phaseR = phaseFunctionR(nu);
	float phaseM = phaseFunctionM(nu);
	vec4 inscatter = max(texture4D(InscatterTexture, r, mu, muS, nu), 0.0);
	
	float r0 = length(x0);
	float mu0 = dot(x0, v) / r0;
	float muS0 = dot(x0, s) / r0;
	
	// Originally there was an offset value (0.01km), but with terrain around elevation 0 this caused problems.
	// I guess the scattering algorithm doesn't work with things "below" the atmosphere.
	// So make sure that all terrain is above 0m ?
	
	//if (r0 > Rg + 0.01) {
	if (r0 > Rg)
	{
		// computes S[L]-T(x,x0)S[L]|x0
		inscatter = max(inscatter - attenuation.rgbr * texture4D(InscatterTexture, r0, mu0, muS0, nu), 0.0);
		
#ifdef FIX
		// avoids imprecision problems near horizon by interpolating between two points above and below horizon
		const float EPS = 0.04; // was 0.004
		float muHoriz = -sqrt(1.0 - (Rg / r) * (Rg / r));
		if (abs(mu - muHoriz) < EPS) {
			float a = ((mu - muHoriz) + EPS) / (2.0 * EPS);

			mu = muHoriz - EPS;
			r0 = sqrt(r * r + t * t + 2.0 * r * t * mu);
			mu0 = (r * mu + t) / r0;
			vec4 inScatter0 = texture4D(InscatterTexture, r, mu, muS, nu);
			vec4 inScatter1 = texture4D(InscatterTexture, r0, mu0, muS0, nu);
			vec4 inScatterA = max(inScatter0 - attenuation.rgbr * inScatter1, 0.0);

			mu = muHoriz + EPS;
			r0 = sqrt(r * r + t * t + 2.0 * r * t * mu);
			mu0 = (r * mu + t) / r0;
			inScatter0 = texture4D(InscatterTexture, r, mu, muS, nu);
			inScatter1 = texture4D(InscatterTexture, r0, mu0, muS0, nu);
			vec4 inScatterB = max(inScatter0 - attenuation.rgbr * inScatter1, 0.0);

			inscatter = mix(inScatterA, inScatterB, a);
		}
#endif
	}
	
	// avoids imprecision problems in Mie scattering when sun is below horizon
	inscatter.w *= smoothstep(0.00, 0.02, muS);
	
	vec3 result = max(inscatter.rgb * phaseR + getMie(inscatter) * phaseM, 0.0);
	
    return result * ISun;
}

// ground radiance at end of ray x+tv, when sun in direction s
// attenuated between ground and viewer (=R[L0]+R[L*])
vec3 ground_radiance(in vec3 x, in vec3 x0, in float n_dot_l, in vec3 reflectance, in vec3 v, in vec3 s, in float mu, in vec3 attenuation) {
	float r0 = length(x0);
	vec3 n = x0 / r0;
	
	// direct sun light (radiance) reaching x0
	float mu0 = dot(n, s);
	vec3 sunLight = transmittanceWithShadow(r0, mu0);
	
	// precomputed sky light (irradiance) (=E[L*]) at x0
	vec3 groundSkyLight = irradiance(IrradianceTexture, r0, mu0);
	
	// light reflected at x0 (=(R[L0]+R[L*])/T(x,x0))
	vec3 groundColor = reflectance * (n_dot_l * max(mu0, 0.0) * sunLight + groundSkyLight) * ISun / M_PI;
	
	return attenuation * groundColor;
}

vec3 HDR(in vec3 L) {
    L = L * exposure;
    L.r = L.r < 1.413 ? pow(L.r * 0.38317, 1.0 / 2.2) : 1.0 - exp(-L.r);
    L.g = L.g < 1.413 ? pow(L.g * 0.38317, 1.0 / 2.2) : 1.0 - exp(-L.g);
    L.b = L.b < 1.413 ? pow(L.b * 0.38317, 1.0 / 2.2) : 1.0 - exp(-L.b);
    return L;
}

layout(location = 0) out vec3 sky_out;

void main() {
	vec2 tex_coords = gl_FragCoord.xy / WindowSize;
	
	if (g_get_object_type(tex_coords) != g_TYPE_TERRAIN) {
		discard;
	}
	
	vec3 offset = vec3(0.0, 0.0, Rg);
	
	// world space camera pos (in km, not m)
	vec3 x = cam_pos.xzy / 1000.0 + offset;
	
	// normalize ray dir
    vec3 v = normalize(ray.xzy);
	
    float r = length(x);
    float mu = dot(x, v) / r;
	
	vec3 s = normalize(sun_dir.xzy);
	
	float depth = g_get_depth(tex_coords);
	vec3 ws_pos = g_get_ws_position(tex_coords, depth, InvProjMatrix, InvModelViewMatrix);
	vec3 x0 = ws_pos.xzy / 1000.0 + offset;
	
	vec3 lit_terrain = texture(InputTexture, tex_coords).rgb;
	vec3 reflectance = g_get_diffuse(tex_coords);
	
	vec3 n = g_get_vs_normal(tex_coords);
	float n_dot_l = max(dot(n, normalize(vs_sun_dir)), 0.0);
	
	float t = distance(x, x0);
	
	vec3 attenuation = attenuation(r, mu, t);
	
	vec3 color = vec3(0.0);
	color += lit_terrain * attenuation;
	color += ground_inscatter(x, x0, t, v, s, r, mu, attenuation); //S[L]-T(x,xs)S[l]|xs
	color += ground_radiance(x, x0, n_dot_l, reflectance * 0.2, v, s, mu, attenuation); // R[L0]+R[L*]
	
	sky_out = HDR(color);
}
