
const vec3 WireColor = vec3(0.1, 0.1, 0.1);
noperspective in vec3 dist;
uniform int ShowWireframe = 0;

in vec2 TexCoords;
in vec3 Position;
in vec2 DepthPos;

uniform vec3 CameraPosition;

uniform mat3 NormalMatrix;
uniform sampler2D NormalsTexture;


uniform float MinHeight;
uniform float MaxHeight;
uniform sampler2DArray AltitudeGradientArray;

float GradientFactor(vec3 normal)
{
	vec3 up = vec3(0.0, 1.0, 0.0);
	float angle = acos(dot(up, normal));
	
	return clamp(angle, 0.0, 1.0);
}

float AltitudeFactor()
{
	return (Position.y - MinHeight) / (MaxHeight - MinHeight);
}


uniform int DetailArraySize;
uniform sampler2DArray DetailAlbedoArray, DetailNormalArray;

// TODO: get these from the app?
const float TextureScale = 1.0 / 40.0;
const float DetailNormalFactor = 0.5;

struct TriplanarValues
{
	vec3 weights;
	vec3 signs;
	vec2[3] coords;
} 
tpValues;

void SetTriplanarValues(vec3 normal)
{
	tpValues.weights = abs(normal);
	tpValues.weights = (tpValues.weights - 0.2) * 7.0;
	tpValues.weights = max(tpValues.weights, vec3(0.0));
	tpValues.weights /= tpValues.weights.x + tpValues.weights.y + tpValues.weights.z;
	
	tpValues.signs = sign(normal);
	
	tpValues.coords[0] = vec2(-tpValues.signs.x * Position.z, Position.y) * TextureScale;
	tpValues.coords[1] = vec2(tpValues.signs.y * Position.x, -Position.z) * TextureScale;
	tpValues.coords[2] = vec2(tpValues.signs.z * Position.x, Position.y) * TextureScale;
}

vec3 GetTriplanarAlbedo(int layer)
{
	vec3 color = vec3(0.0);
	
	for (int i = 0; i != 3; ++i)
	{
		vec3 coords = vec3(tpValues.coords[i], layer);
		color += tpValues.weights[i] * texture(DetailAlbedoArray, coords).rgb;
	}
	
	return color;
}

vec3 GetTriplanarNormal(int layer)
{
	vec3[3] normal;
	
	for (int i = 0; i != 3; ++i)
	{
		vec3 coords = vec3(tpValues.coords[i], layer);
		normal[i] = tpValues.weights[i] * normalize(texture(DetailNormalArray, coords).rgb * 2.0 - 1.0);
	}
	
	// Alter the normal for the correct plane (hacky...)
	vec3 bump = vec3(0.0);	
	bump += vec3(0.0, normal[0].y, -tpValues.signs.x * normal[0].x);
	bump += vec3(tpValues.signs.y * normal[1].x, 0.0, -normal[1].y);
	bump += vec3(tpValues.signs.z * normal[2].x, normal[2].y, 0.0);
	
	return bump;
}


void GetDetails(float gradient, float altitude, vec3 normal, out vec3 detailAlbedo, out vec3 detailNormal)
{
	detailAlbedo = vec3(0.0);
	detailNormal = vec3(0.0);
	
	SetTriplanarValues(normal);
	
	float alphaAccum = 0.0;
	
	for (int i = 0; i != DetailArraySize; ++i)
	{
		float altGrad = texture(AltitudeGradientArray, vec3(gradient, altitude, i)).r;
		
		float alpha = min(altGrad, 1.0 - alphaAccum);
		
		detailAlbedo += GetTriplanarAlbedo(i) * alpha;
		detailNormal += GetTriplanarNormal(i) * alpha;
		
		alphaAccum += alpha;
	}
	
	// Fill the remaining alpha space (is this actually needed + useful?)
	//float alphaRemaining = 1.0 - alphaAccum;
	//
	//detailAlbedo += GetTriplanarAlbedo(0) * alphaRemaining;
	//detailNormal += GetTriplanarNormal(0) * alphaRemaining;
}

// from http://www.thetenthplanet.de/archives/1180
mat3 GetCotangentFrame(const in vec3 normal)
{
	vec3 viewDir = CameraPosition - Position;
	
	vec3 dpx = dFdx(viewDir);
	vec3 dpy = dFdy(viewDir);
	
	vec2 dtx = dFdx(TexCoords);
	vec2 dty = dFdy(TexCoords);
	
    vec3 dpyperp = cross(dpy, normal);
    vec3 dpxperp = cross(normal, dpx);
    
    vec3 tangent =  dpyperp * dtx.x + dpxperp * dty.x;
    vec3 binormal = dpyperp * dtx.y + dpxperp * dty.y;
 	
    // construct a scale-invariant frame
    float invmax = inversesqrt(max(dot(tangent,tangent), dot(binormal,binormal)));
    
    return mat3(tangent * invmax, binormal * invmax, normal);
	
}

vec3 PerturbNormal(vec3 normal, vec3 detailNormal)
{
    mat3 TBN = GetCotangentFrame(normal);
    
    return normalize(TBN * detailNormal);
}


vec3 AddWireframe(vec3 detailAlbedo)
{
	float d = min(dist[0], min(dist[1], dist[2]));
 	float I = exp2(-2 * d * d);
 	
	return mix(detailAlbedo, WireColor, I * ShowWireframe);
}


void main()
{
	vec3 normal = normalize(texture(NormalsTexture, TexCoords).xyz);
	
	float gradient = GradientFactor(normal);
	float altitude = AltitudeFactor();
	
	vec3 detailAlbedo, detailNormal;
	GetDetails(gradient, altitude, normal, detailAlbedo, detailNormal);
	
	vec3 perturbedNormal = normalize(normal + detailNormal * DetailNormalFactor);
	
	// output to g_buffer
	g_set_diffuse(AddWireframe(detailAlbedo));
	g_set_normal(NormalMatrix * perturbedNormal);
	g_set_depth(DepthPos.x / DepthPos.y);
	g_set_object_type(g_TYPE_TERRAIN);
}
