-- Contains code preventing the use of globals without previous declaration
-- MUST be parsed before any other lua script

-- keep hold of the file currently being parsed
local current_file = "none"
-- all global variable names / statuses kept according to file
local globals_by_file = {
	["none"] = {} -- C++ defined stuff goes here.
}
-- status of globals
local status = {
	DECLARED = 0, -- declared, but not yet defined
	DEFINED = 1, -- declared and defined ok
	UNDEFINED = 2, -- declared, but not defined, so marked as undefined
}

-- squirrel away the old dofile for use by artefract_dofile
local dofile = dofile

-- new dofile to make sure globals were properly declared and defined
local function artefract_dofile(f)
	local ocf = current_file
	local stem = ocf:match("^(.+[/\\]).-$") or ""
	
	-- try file relative to current_file
	current_file = stem..f
	globals_by_file[current_file] = globals_by_file[current_file] or {}
	if dofile(current_file) then
		
		-- try file relative to root dir
		current_file = f
		globals_by_file[current_file] = globals_by_file[current_file] or {}
		if dofile(current_file) then
			error("failed to parse lua file: "..current_file)
		end
	end
	
	for k,v in pairs(globals_by_file[current_file]) do
		-- also check that it's nil, since
		if v == status.DECLARED and rawget(_G, k) == nil then
			warning("global variable: "..k
				.." declared, but never defined in file: "..current_file)
			globals_by_file[current_file].v = status.UNDEFINED
		end
	end
	
	current_file = ocf
end

-- make dofile call artefract_dofile from now on
_G.dofile = artefract_dofile

-- use declare_globals with a table of strings to declare new global variables
local function declare_globals(t)
	if type(t) ~= "table" then
		warning("function declare_globals' must be called with a table.")
		return
	end
	for _,v in ipairs(t) do
		if globals_by_file[current_file][v] then
			warning("redeclaration of global variable: "..v.." in file: "..current_file)
		end
		globals_by_file[current_file][v] = status.DECLARED
		-- would prevent simple(ish) redefinition of already defined globals (e.g. error)
		-- rawset(_G, v, nil)
	end
end

-- helper function for metatable stuff below
local function find_global(n)
	for fk,fv in pairs(globals_by_file) do
		for nk,nv in pairs(fv) do
			if nk == n then
				return fk, nv
			end
		end
	end
end

-- actually implements the restrictions by checking the globals list
setmetatable(_G, {
	__newindex = function(t, n, v)
		local fk, nv = find_global(n)
		if nv == status.DECLARED then
			globals_by_file[fk][n] = status.DEFINED
			rawset(t, n, v)
			return
		end
		local info = debug.getinfo(2, "S")
		if not info or info.what == "C" then
			globals_by_file["none"][n] = status.DEFINED
			rawset(t, n, v)
			return
		end
		error("attempt to assign to an undeclared global variable: "
			..n.." in file: "..current_file, 2)
	end,
	
	__index = function(_, n)
		if not find_global(n) then
			error("attempt to read from an undeclared global variable: "
				..n.." in file: "..current_file, 2)
		end
		return nil
	end,
})

declare_globals{
	"artefract_dofile",
	"declare_globals",
}

_G.artefract_dofile = artefract_dofile
_G.declare_globals = declare_globals
